from deltacat.types.media import DistributedDatasetType
from deltacat.compute.merge_on_read.daft import merge as daft_merge

MERGE_FUNC_BY_DISTRIBUTED_DATASET_TYPE = {DistributedDatasetType.DAFT.value: daft_merge}



import unittest
import sqlite3
import ray
import os
import deltacat.tests.local_deltacat_storage as ds
from deltacat.utils.common import current_time_ms
from deltacat.tests.test_utils.pyarrow import (
    create_delta_from_csv_file,
    commit_delta_to_partition,
)
from deltacat.types.media import DistributedDatasetType, ContentType
from deltacat.catalog import default_catalog_impl as dc


class TestReadTable(unittest.TestCase):
    READ_TABLE_NAMESPACE = "catalog_read_table_namespace"
    LOCAL_CATALOG_NAME = "local_catalog"
    DB_FILE_PATH = f"{current_time_ms()}.db"
    SAMPLE_FILE_PATH = "deltacat/tests/catalog/data/sample_table.csv"

    @classmethod
    def setUpClass(cls):
        ray.init(local_mode=True, ignore_reinit_error=True)

        con = sqlite3.connect(cls.DB_FILE_PATH)
        cur = con.cursor()
        cls.kwargs = {
            ds.SQLITE_CON_ARG: con,
            ds.SQLITE_CUR_ARG: cur,
            "supported_content_types": [ContentType.CSV],
        }
        cls.deltacat_storage_kwargs = {ds.DB_FILE_PATH_ARG: cls.DB_FILE_PATH}

        super().setUpClass()

    @classmethod
    def doClassCleanups(cls) -> None:
        os.remove(cls.DB_FILE_PATH)
        ray.shutdown()
        super().tearDownClass()

    def test_daft_distributed_read_sanity(self):
        # setup
        READ_TABLE_TABLE_NAME = "test_read_table"
        create_delta_from_csv_file(
            self.READ_TABLE_NAMESPACE,
            [self.SAMPLE_FILE_PATH],
            table_name=READ_TABLE_TABLE_NAME,
            **self.kwargs,
        )

        dc.initialize(ds=ds)
        df = dc.read_table(
            table=READ_TABLE_TABLE_NAME,
            namespace=self.READ_TABLE_NAMESPACE,
            catalog=self.LOCAL_CATALOG_NAME,
            distributed_dataset_type=DistributedDatasetType.DAFT,
            deltacat_storage_kwargs=self.kwargs,
        )

        # verify
        self.assertEqual(df.count_rows(), 6)
        self.assertEqual(df.column_names, ["pk", "value"])

    def test_daft_distributed_read_multiple_deltas(self):
        # setup
        READ_TABLE_TABLE_NAME = "test_read_table_2"
        delta = create_delta_from_csv_file(
            self.READ_TABLE_NAMESPACE,
            [self.SAMPLE_FILE_PATH],
            table_name=READ_TABLE_TABLE_NAME,
            **self.kwargs,
        )

        partition = ds.get_partition(
            delta.stream_locator, delta.partition_values, **self.kwargs
        )

        commit_delta_to_partition(
            partition=partition, file_paths=[self.SAMPLE_FILE_PATH], **self.kwargs
        )

        # action
        dc.initialize(ds=ds)
        df = dc.read_table(
            table=READ_TABLE_TABLE_NAME,
            namespace=self.READ_TABLE_NAMESPACE,
            catalog=self.LOCAL_CATALOG_NAME,
            distributed_dataset_type=DistributedDatasetType.DAFT,
            merge_on_read=False,
            deltacat_storage_kwargs=self.kwargs,
        )

        # verify
        self.assertEqual(
            df.count_rows(),
            12,
            "we expect twice as many" " columns as merge on read is disabled",
        )
        self.assertEqual(df.column_names, ["pk", "value"])



from __future__ import annotations
from typing import List
from enum import Enum


class TransformName(str, Enum):
    IDENTITY = "identity"
    BUCKET = "bucket"


class TransformParameters(dict):
    """
    This is a parent class that contains properties
    to be passed to the corresponding transform
    """

    pass


class IdentityTransformParameters(TransformParameters):
    """
    This class is used to pass parameters to the identity transform
    """

    @staticmethod
    def of(column_name: str) -> IdentityTransformParameters:
        identify_transform_parameters = IdentityTransformParameters()
        identify_transform_parameters["columnName"] = column_name
        return identify_transform_parameters

    @property
    def column_name(self) -> str:
        """
        The name of the column to use for identity transform
        """
        return self["columnName"]

    @column_name.setter
    def column_name(self, value: str) -> None:
        self["columnName"] = value


class BucketingStrategy(str, Enum):
    """
    A bucketing strategy for the transform
    """

    # Uses default deltacat bucketing strategy.
    # This strategy supports hashing on composite keys
    # and uses SHA1 hashing for determining the bucket.
    # If no columns passed, it will use a random UUID
    # for determining the bucket.
    DEFAULT = "default"

    # Uses iceberg compliant bucketing strategy.
    # As indicated in the iceberg spec, it does not support
    # composite keys and uses murmur3 hash for determining
    # the bucket.
    # See https://iceberg.apache.org/spec/#partitioning
    ICEBERG = "iceberg"


class BucketTransformParameters(TransformParameters):
    """
    Encapsulates parameters for the bucket transform.
    """

    def of(
        self,
        num_buckets: int,
        column_names: List[str],
        bucketing_strategy: BucketingStrategy,
    ) -> BucketTransformParameters:
        bucket_transform_parameters = BucketTransformParameters()
        bucket_transform_parameters["numBuckets"] = num_buckets
        bucket_transform_parameters["columnNames"] = column_names
        bucket_transform_parameters["bucketingStrategy"] = bucketing_strategy

        return bucket_transform_parameters

    @property
    def num_buckets(self) -> int:
        """
        The total number of buckets to create for values of the column
        """
        return self["numBuckets"]

    @property
    def column_names(self) -> List[str]:
        """
        An ordered list of unique column names from the table schema
        to use for bucketings.
        """
        return self["columnNames"]

    @property
    def bucketing_strategy(self) -> BucketingStrategy:
        """
        The bucketing strategy to used.
        """
        return self["bucketingStrategy"]


class Transform(dict):
    """
    A transform is represents how a particular column value can be
    transformed into a new value. This is mostly used in the context
    of partitioning the data files in a table.
    """

    @staticmethod
    def of(
        name: TransformName,
        parameters: TransformParameters,
    ) -> Transform:
        partition_transform = Transform()
        partition_transform["name"] = name
        partition_transform["parameters"] = parameters
        return partition_transform

    @property
    def name(self) -> TransformName:
        return self["name"]

    @property
    def parameters(self) -> TransformParameters:
        return self["parameters"]



import botocore
from typing import Set
from daft.exceptions import DaftTransientError
from deltacat.utils.common import env_integer, env_string


DAFT_MAX_S3_CONNECTIONS_PER_FILE = env_integer("DAFT_MAX_S3_CONNECTIONS_PER_FILE", 8)
DEFAULT_FILE_READ_TIMEOUT_MS = env_integer(
    "DEFAULT_FILE_READ_TIMEOUT_MS", 300_000
)  # 5 mins
BOTO_MAX_RETRIES = env_integer("BOTO_MAX_RETRIES", 5)
BOTO_TIMEOUT_ERROR_CODES: Set[str] = {"ReadTimeoutError", "ConnectTimeoutError"}
BOTO_THROTTLING_ERROR_CODES: Set[str] = {"Throttling", "SlowDown"}
RETRYABLE_TRANSIENT_ERRORS = (
    OSError,
    botocore.exceptions.ConnectionError,
    botocore.exceptions.HTTPClientError,
    botocore.exceptions.NoCredentialsError,
    botocore.exceptions.ConnectTimeoutError,
    botocore.exceptions.ReadTimeoutError,
    DaftTransientError,
)
AWS_REGION = env_string("AWS_REGION", "us-east-1")
UPLOAD_DOWNLOAD_RETRY_STOP_AFTER_DELAY = env_integer(
    "UPLOAD_DOWNLOAD_RETRY_STOP_AFTER_DELAY", 10 * 60
)
UPLOAD_SLICED_TABLE_RETRY_STOP_AFTER_DELAY = env_integer(
    "UPLOAD_SLICED_TABLE_RETRY_STOP_AFTER_DELAY", 30 * 60
)
DOWNLOAD_MANIFEST_ENTRY_RETRY_STOP_AFTER_DELAY = env_integer(
    "DOWNLOAD_MANIFEST_ENTRY_RETRY_STOP_AFTER_DELAY", 30 * 60
)



from __future__ import annotations
from typing import Optional, Dict, List, Union, Any
from deltacat.storage import (
    Delta,
    DeltaLocator,
    interface as unimplemented_deltacat_storage,
)


class MergeOnReadParams(dict):
    """
    This class represents the parameters passed to compact_partition (deltacat/compute/compactor/compaction_session.py)
    """

    @staticmethod
    def of(params: Optional[Dict]) -> MergeOnReadParams:
        params = {} if params is None else params

        result = MergeOnReadParams(params)
        assert result.deltas is not None, "deltas is a required arg"

        result.deltacat_storage = params.get(
            "deltacat_storage", unimplemented_deltacat_storage
        )
        result.reader_kwargs = params.get("reader_kwargs", {})
        result.deltacat_storage_kwargs = params.get("deltacat_storage_kwargs", {})

        return result

    @property
    def deltas(self) -> List[Union[Delta, DeltaLocator]]:
        """
        The list of deltas to compact in-memory.
        """
        return self["deltas"]

    @deltas.setter
    def deltas(self, to_set: List[Union[Delta, DeltaLocator]]) -> None:
        self["deltas"] = to_set

    @property
    def reader_kwargs(self) -> Dict[Any, Any]:
        """
        The key word arguments to be passed to the reader.
        """
        return self["reader_kwargs"]

    @reader_kwargs.setter
    def reader_kwargs(self, kwargs: Dict[Any, Any]) -> None:
        self["reader_kwargs"] = kwargs

    @property
    def deltacat_storage(self) -> unimplemented_deltacat_storage:
        return self["deltacat_storage"]

    @deltacat_storage.setter
    def deltacat_storage(self, storage: unimplemented_deltacat_storage) -> None:
        self["deltacat_storage"] = storage

    @property
    def deltacat_storage_kwargs(self) -> dict:
        return self["deltacat_storage_kwargs"]

    @deltacat_storage_kwargs.setter
    def deltacat_storage_kwargs(self, kwargs: dict) -> None:
        self["deltacat_storage_kwargs"] = kwargs
